@@ -87,6 +87,7 @@ typedef struct {
 	buffer *cgi_handler;      /* dumb pointer */
 	http_response_opts opts;
 	plugin_config conf;
+	off_t orig_reqbody_length;
 } handler_ctx;
 
 typedef struct cgi_pid_t {
@@ -98,14 +99,10 @@ typedef struct cgi_pid_t {
 } cgi_pid_t;
 
 static handler_ctx * cgi_handler_ctx_init(void) {
-	handler_ctx *hctx = calloc(1, sizeof(*hctx));
-
-	force_assert(hctx);
-
+	handler_ctx *hctx = ck_calloc(1, sizeof(*hctx));
 	hctx->response = chunk_buffer_acquire();
 	hctx->fd = -1;
 	hctx->fdtocgi = -1;
-
 	return hctx;
 }
 
@@ -115,22 +112,18 @@ static void cgi_handler_ctx_free(handler_ctx *hctx) {
 }
 
 INIT_FUNC(mod_cgi_init) {
-	plugin_data *p;
+	plugin_data * const p = ck_calloc(1, sizeof(*p));
 	const char *s;
 
-	p = calloc(1, sizeof(*p));
-
-	force_assert(p);
-
 	/* for valgrind */
 	s = getenv("LD_PRELOAD");
-	if (s) p->env.ld_preload = buffer_init_string(s);
+	if (s) buffer_copy_string((p->env.ld_preload = buffer_init()), s);
 	s = getenv("LD_LIBRARY_PATH");
-	if (s) p->env.ld_library_path = buffer_init_string(s);
+	if (s) buffer_copy_string((p->env.ld_library_path = buffer_init()), s);
       #ifdef __CYGWIN__
 	/* CYGWIN needs SYSTEMROOT */
 	s = getenv("SYSTEMROOT");
-	if (s) p->env.systemroot = buffer_init_string(s);
+	if (s) buffer_copy_string((p->env.systemroot = buffer_init()), s);
       #endif
 
 	return p;
@@ -270,8 +263,7 @@ static int mod_cgi_str_to_signal (const char *s, int default_sig) {
 }
 
 static cgi_limits * mod_cgi_parse_limits(const array * const a, log_error_st * const errh) {
-    cgi_limits * const limits = calloc(1, sizeof(cgi_limits));
-    force_assert(limits);
+    cgi_limits * const limits = ck_calloc(1, sizeof(cgi_limits));
     for (uint32_t i = 0; i < a->used; ++i) {
         const data_unset * const du = a->data[i];
         int32_t v = config_plugin_value_to_int32(du, -1);
@@ -384,8 +376,7 @@ SETDEFAULTS_FUNC(mod_cgi_set_defaults) {
 
 
 static cgi_pid_t * cgi_pid_add(plugin_data *p, pid_t pid, handler_ctx *hctx) {
-    cgi_pid_t *cgi_pid = malloc(sizeof(cgi_pid_t));
-    force_assert(cgi_pid);
+    cgi_pid_t *cgi_pid = ck_malloc(sizeof(cgi_pid_t));
     cgi_pid->pid = pid;
     cgi_pid->signal_sent = 0;
     cgi_pid->hctx = hctx;
@@ -415,6 +406,7 @@ static void cgi_pid_del(plugin_data *p, cgi_pid_t *cgi_pid) {
 
 static void cgi_connection_close_fdtocgi(handler_ctx *hctx) {
 	/*(closes only hctx->fdtocgi)*/
+	if (-1 == hctx->fdtocgi) return;
 	struct fdevents * const ev = hctx->ev;
 	fdevent_fdnode_event_del(ev, hctx->fdntocgi);
 	/*fdevent_unregister(ev, hctx->fdtocgi);*//*(handled below)*/
@@ -521,6 +513,13 @@ static handler_t cgi_response_headers(request_st * const r, struct http_response
     if (light_btst(r->resp_htags, HTTP_HEADER_UPGRADE)) {
         if (hctx->conf.upgrade && r->http_status == 101) {
             /* 101 Switching Protocols; transition to transparent proxy */
+            if (r->h2_connect_ext) {
+                r->http_status = 200; /* OK (response status for CONNECT) */
+                http_header_response_unset(r, HTTP_HEADER_UPGRADE,
+                                           CONST_STR_LEN("Upgrade"));
+                http_header_response_unset(r, HTTP_HEADER_OTHER,
+                                         CONST_STR_LEN("Sec-WebSocket-Accept"));
+            }
             http_response_upgrade_read_body_unknown(r);
         }
         else {
@@ -534,11 +533,20 @@ static handler_t cgi_response_headers(request_st * const r, struct http_response
           #endif
         }
     }
+    else if (__builtin_expect( (r->h2_connect_ext != 0), 0)
+             && r->http_status < 300) {
+        /*(not handling other 1xx intermediate responses here; not expected)*/
+        http_response_body_clear(r, 0);
+        r->handler_module = NULL;
+        r->http_status = 405; /* Method Not Allowed */
+        return HANDLER_FINISHED;
+    }
 
     if (hctx->conf.upgrade
         && !light_btst(r->resp_htags, HTTP_HEADER_UPGRADE)) {
         chunkqueue *cq = &r->reqbody_queue;
         hctx->conf.upgrade = 0;
+        r->reqbody_length = hctx->orig_reqbody_length;
         if (cq->bytes_out == (off_t)r->reqbody_length) {
             cgi_connection_close_fdtocgi(hctx); /*(closes hctx->fdtocgi)*/
         }
@@ -554,7 +562,7 @@ static handler_t cgi_local_redir(request_st * const r, handler_ctx * const hctx)
     buffer_clear(hctx->response);
     chunk_buffer_yield(hctx->response);
     http_response_reset(r); /*(includes r->http_status = 0)*/
-    plugins_call_handle_request_reset(r);
+    r->con->srv->plugins_request_reset(r);
     /*cgi_connection_close(hctx);*//*(already cleaned up and hctx is now invalid)*/
     return HANDLER_COMEBACK;
 }
@@ -621,6 +629,7 @@ static handler_t cgi_process_rd_revents(handler_ctx * const hctx, request_st * c
 			if (0 == r->http_status) r->http_status = 200; /* OK */
 		}
 		cgi_connection_close(hctx);
+		return HANDLER_FINISHED;
 	} else if (revents & FDEVENT_ERR) {
 		/* kill all connections to the cgi process */
 		cgi_connection_close(hctx);
@@ -767,7 +776,8 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 		}
 	}
 	else if (!(r->conf.stream_request_body /*(if not streaming request body)*/
-	           & (FDEVENT_STREAM_REQUEST|FDEVENT_STREAM_REQUEST_BUFMIN))) {
+	           & (FDEVENT_STREAM_REQUEST|FDEVENT_STREAM_REQUEST_BUFMIN))
+	         && !hctx->conf.upgrade) {
 		chunkqueue * const cq = &r->reqbody_queue;
 		chunk * const c = cq->first;
 		if (c && c == cq->last && c->type == FILE_CHUNK && c->file.is_temp) {
@@ -791,7 +801,7 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 
 	unsigned int bufsz_hint = 16384;
   #ifdef _WIN32
-	if (r->reqbody_length <= 1048576)
+	if (r->reqbody_length <= 1048576 && r->reqbody_length > 0)
 		bufsz_hint = (unsigned int)r->reqbody_length;
   #endif
 	if (-1 == to_cgi_fds[0] && fdevent_pipe_cloexec(to_cgi_fds, bufsz_hint)) {
@@ -824,8 +834,42 @@ static int cgi_create_env(request_st * const r, plugin_data * const p, handler_c
 
 		/* create environment */
 
+		if (r->h2_connect_ext) {
+			/*(SERVER_PROTOCOL=HTTP/1.1 instead of HTTP/2.0)*/
+			r->http_version = HTTP_VERSION_1_1;
+			r->http_method = HTTP_METHOD_GET;
+		}
+		if (hctx->conf.upgrade) {
+			r->reqbody_length = hctx->orig_reqbody_length;
+			if (r->reqbody_length < 0)
+				r->reqbody_length = 0;
+		}
+
 		http_cgi_headers(r, &opts, cgi_env_add, env);
 
+		if (hctx->conf.upgrade)
+			r->reqbody_length = -1;
+		if (r->h2_connect_ext) {
+			r->http_version = HTTP_VERSION_2;
+			r->http_method = HTTP_METHOD_CONNECT;
+			/* https://datatracker.ietf.org/doc/html/rfc6455#section-4.1
+			 * 7. The request MUST include a header field with the name
+			 *    |Sec-WebSocket-Key|.  The value of this header field MUST be a
+			 *    nonce consisting of a randomly selected 16-byte value that has
+			 *    been base64-encoded (see Section 4 of [RFC4648]).  The nonce
+			 *    MUST be selected randomly for each connection.
+			 * Note: Sec-WebSocket-Key is not used in RFC8441;
+			 *       include Sec-WebSocket-Key for HTTP/1.1 compatibility;
+			 *       !!not random!! base64-encoded "0000000000000000" */
+			if (!http_header_request_get(r, HTTP_HEADER_OTHER,
+			                             CONST_STR_LEN("Sec-WebSocket-Key")))
+				cgi_env_add(env, CONST_STR_LEN("HTTP_SEC_WEBSOCKET_KEY"),
+				                 CONST_STR_LEN("MDAwMDAwMDAwMDAwMDAwMA=="));
+			/*(Upgrade and Connection should not exist for HTTP/2 request)*/
+			cgi_env_add(env, CONST_STR_LEN("HTTP_UPGRADE"), CONST_STR_LEN("websocket"));
+			cgi_env_add(env, CONST_STR_LEN("HTTP_CONNECTION"), CONST_STR_LEN("upgrade"));
+		}
+
 		/* for valgrind */
 		if (p->env.ld_preload) {
 			cgi_env_add(env, CONST_STR_LEN("LD_PRELOAD"), BUF_PTR_LEN(p->env.ld_preload));
@@ -964,12 +1008,18 @@ URIHANDLER_FUNC(cgi_is_handled) {
 	if (!S_ISREG(st->st_mode)) return HANDLER_GO_ON;
 	if (p->conf.execute_x_only == 1 && (st->st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) == 0) return HANDLER_GO_ON;
 
+	if (__builtin_expect( (r->h2_connect_ext != 0), 0) && !p->conf.upgrade) {
+		r->http_status = 405; /* Method Not Allowed */
+		return HANDLER_FINISHED;
+	}
+
 	if (r->reqbody_length
 	    && p->tempfile_accum
 	    && !(r->conf.stream_request_body /*(if not streaming request body)*/
 	         & (FDEVENT_STREAM_REQUEST|FDEVENT_STREAM_REQUEST_BUFMIN))) {
 		/* store request body in single tempfile if not streaming request body*/
-		r->reqbody_queue.upload_temp_file_size = INTMAX_MAX;
+		r->reqbody_queue.upload_temp_file_size =
+		  (off_t)((1uLL << (sizeof(off_t)*8-1))-1);
 	}
 
 	{
@@ -980,10 +1030,19 @@ URIHANDLER_FUNC(cgi_is_handled) {
 		hctx->plugin_data = p;
 		hctx->cgi_handler = &ds->value;
 		memcpy(&hctx->conf, &p->conf, sizeof(plugin_config));
-		hctx->conf.upgrade =
-		  hctx->conf.upgrade
-		  && r->http_version == HTTP_VERSION_1_1
-		  && light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE);
+		if (__builtin_expect( (r->h2_connect_ext != 0), 0)) {
+		}
+		else if (!light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE))
+			hctx->conf.upgrade = 0;
+		else if (!hctx->conf.upgrade || r->http_version != HTTP_VERSION_1_1) {
+			hctx->conf.upgrade = 0;
+			http_header_request_unset(r, HTTP_HEADER_UPGRADE,
+			                          CONST_STR_LEN("Upgrade"));
+		}
+		if (hctx->conf.upgrade) {
+			hctx->orig_reqbody_length = r->reqbody_length;
+			r->reqbody_length = -1;
+		}
 		hctx->opts.max_per_read =
 		  !(r->conf.stream_response_body /*(if not streaming response body)*/
 		    & (FDEVENT_STREAM_RESPONSE|FDEVENT_STREAM_RESPONSE_BUFMIN))
@@ -1060,29 +1119,25 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 		if (chunkqueue_length(cq) > 65536 - 4096
 		    && (r->conf.stream_request_body & FDEVENT_STREAM_REQUEST_BUFMIN)){
 			r->conf.stream_request_body &= ~FDEVENT_STREAM_REQUEST_POLLIN;
-			if (-1 != hctx->fd) return HANDLER_WAIT_FOR_EVENT;
 		} else {
 			handler_t rc = r->con->reqbody_read(r);
-			if (!chunkqueue_is_empty(cq)) {
-				if (fdevent_fdnode_interest(hctx->fdntocgi) & FDEVENT_OUT) {
-					return (rc == HANDLER_GO_ON) ? HANDLER_WAIT_FOR_EVENT : rc;
-				}
-			}
-			if (rc != HANDLER_GO_ON) return rc;
-
-			/* CGI environment requires that Content-Length be set.
-			 * Send 411 Length Required if Content-Length missing.
-			 * (occurs here if client sends Transfer-Encoding: chunked
-			 *  and module is flagged to stream request body to backend) */
-			if (-1 == r->reqbody_length) {
-				return (r->conf.stream_request_body & FDEVENT_STREAM_REQUEST)
-				  ? http_response_reqbody_read_error(r, 411)
-				  : HANDLER_WAIT_FOR_EVENT;
-			}
+			if (rc != HANDLER_GO_ON
+			    && !(hctx->conf.upgrade && -1 == hctx->fd
+			         && rc == HANDLER_WAIT_FOR_EVENT))
+				return rc;
 		}
 	}
 
 	if (-1 == hctx->fd) {
+			/* CGI environment requires that Content-Length be set.
+			 * Send 411 Length Required if Content-Length missing.
+			 * (occurs here if client sends Transfer-Encoding: chunked
+			 *  and module is flagged to stream request body to backend) */
+			if (-1 == r->reqbody_length && !hctx->conf.upgrade) {
+				return (r->conf.stream_request_body & FDEVENT_STREAM_REQUEST)
+				  ? http_response_reqbody_read_error(r, 411)
+				  : HANDLER_WAIT_FOR_EVENT;
+			}
 		if (cgi_create_env(r, p, hctx, hctx->cgi_handler)) {
 			r->http_status = 500;
 			r->handler_module = NULL;
@@ -1090,6 +1145,8 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 			return HANDLER_FINISHED;
 		}
 	} else if (!chunkqueue_is_empty(cq)) {
+		if (fdevent_fdnode_interest(hctx->fdntocgi) & FDEVENT_OUT)
+			return HANDLER_WAIT_FOR_EVENT;
 		if (0 != cgi_write_request(hctx, hctx->fdtocgi)) {
 			cgi_connection_close(hctx);
 			return HANDLER_ERROR;
@@ -1187,6 +1244,7 @@ static handler_t cgi_waitpid_cb(server *srv, void *p_d, pid_t pid, int status) {
 }
 
 
+__attribute_cold__
 int mod_cgi_plugin_init(plugin *p);
 int mod_cgi_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
