@@ -1,5 +1,6 @@
 #include "first.h"
 
+#include "plugins.h"
 #include "plugin.h"
 #include "base.h"
 #include "array.h"
@@ -8,8 +9,6 @@
 #include <string.h>
 #include <stdlib.h>
 
-array plugin_stats; /* global */
-
 #ifdef HAVE_VALGRIND_VALGRIND_H
 # include <valgrind/valgrind.h>
 #endif
@@ -56,12 +55,7 @@ typedef enum {
 __attribute_malloc__
 __attribute_returns_nonnull__
 static plugin *plugin_init(void) {
-	plugin *p;
-
-	p = calloc(1, sizeof(*p));
-	force_assert(NULL != p);
-
-	return p;
+	return ck_calloc(1, sizeof(plugin));
 }
 
 static void plugin_free(plugin *p) {
@@ -71,7 +65,7 @@ static void plugin_free(plugin *p) {
      #if defined(HAVE_VALGRIND_VALGRIND_H)
      /*if (!RUNNING_ON_VALGRIND) */
      #endif
-      #if defined(__WIN32)
+      #ifdef _WIN32
         FreeLibrary(p->lib);
       #else
         dlclose(p->lib);
@@ -82,17 +76,18 @@ static void plugin_free(plugin *p) {
     free(p);
 }
 
-static void plugins_register(server *srv, plugin *p) {
-	plugin **ps;
-	if (srv->plugins.used == srv->plugins.size) {
-		srv->plugins.size += 4;
-		srv->plugins.ptr   = realloc(srv->plugins.ptr, srv->plugins.size * sizeof(*ps));
-		force_assert(NULL != srv->plugins.ptr);
-	}
-
-	ps = srv->plugins.ptr;
-	ps[srv->plugins.used++] = p;
+#ifdef _WIN32
+__attribute_cold__
+static void
+log_w32_syserror_2 (log_error_st *const errh, const char *file, const int line, const char * const str1, const char * const str2)
+{
+    TCHAR lpMsgBuf[1024];
+    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
+                  0, /* MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) */
+                  (LPTSTR)lpMsgBuf, sizeof(lpMsgBuf)/sizeof(TCHAR), NULL);
+    log_error(errh, file, line, "%s for %s: %s", str1, str2, (char *)lpMsgBuf);
 }
+#endif
 
 /**
  *
@@ -128,6 +123,9 @@ static const plugin_load_functions load_functions[] = {
 };
 
 int plugins_load(server *srv) {
+	ck_realloc_u32(&srv->plugins.ptr, 0,
+	               srv->srvconf.modules->used, sizeof(plugin *));
+
 	for (uint32_t i = 0; i < srv->srvconf.modules->used; ++i) {
 		data_string *ds = (data_string *)srv->srvconf.modules->data[i];
 		char *module = ds->value.ptr;
@@ -141,7 +139,7 @@ int plugins_load(server *srv) {
 					plugin_free(p);
 					return -1;
 				}
-				plugins_register(srv, p);
+				((plugin **)srv->plugins.ptr)[srv->plugins.used++] = p;
 				break;
 			}
 		}
@@ -157,93 +155,74 @@ int plugins_load(server *srv) {
 
 	return 0;
 }
+
 #else /* defined(LIGHTTPD_STATIC) */
+
 int plugins_load(server *srv) {
+	ck_realloc_u32(&srv->plugins.ptr, 0,
+	               srv->srvconf.modules->used, sizeof(plugin *));
+
 	buffer * const tb = srv->tmp_buf;
-	plugin *p;
 	int (*init)(plugin *pl);
 
 	for (uint32_t i = 0; i < srv->srvconf.modules->used; ++i) {
 		const buffer * const module = &((data_string *)srv->srvconf.modules->data[i])->value;
+		void *lib = NULL;
+
+		/* check if module is built-in to main executable */
+		buffer_clear(tb);
+		buffer_append_str2(tb, BUF_PTR_LEN(module),
+		                       CONST_STR_LEN("_plugin_init"));
+	  #ifdef _WIN32
+		init = (int(WINAPI *)(plugin *))(intptr_t)
+		  GetProcAddress(GetModuleHandle(NULL), tb->ptr);
+	  #else
+		init = (int (*)(plugin *))(intptr_t)dlsym(RTLD_DEFAULT, tb->ptr);
+	  #endif
+
+	  if (NULL == init) {
 		buffer_copy_string(tb, srv->srvconf.modules_dir);
 		buffer_append_path_len(tb, BUF_PTR_LEN(module));
-#if defined(__WIN32) || defined(__CYGWIN__)
+
+	  #ifdef _WIN32
 		buffer_append_string_len(tb, CONST_STR_LEN(".dll"));
-#else
-		buffer_append_string_len(tb, CONST_STR_LEN(".so"));
-#endif
-
-		p = plugin_init();
-#ifdef __WIN32
-		if (NULL == (p->lib = LoadLibrary(tb->ptr))) {
-			LPVOID lpMsgBuf;
-			FormatMessage(
-				FORMAT_MESSAGE_ALLOCATE_BUFFER |
-					FORMAT_MESSAGE_FROM_SYSTEM,
-				NULL,
-				GetLastError(),
-				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-				(LPTSTR) &lpMsgBuf,
-				0, NULL);
-
-			log_error(srv->errh, __FILE__, __LINE__,
-			  "LoadLibrary() failed %s %s", lpMsgBuf, tb->ptr);
-
-			plugin_free(p);
-
+		if (NULL == (lib = LoadLibrary(tb->ptr))) {
+			log_w32_syserror_2(srv->errh, __FILE__, __LINE__,
+			                   "LoadLibrary()", tb->ptr);
 			if (srv->srvconf.compat_module_load) {
 				if (buffer_eq_slen(module, CONST_STR_LEN("mod_deflate")))
 					continue;
 			}
 			return -1;
-
 		}
-#else
-		if (NULL == (p->lib = dlopen(tb->ptr, RTLD_NOW|RTLD_GLOBAL))) {
+		buffer_copy_buffer(tb, module);
+		buffer_append_string_len(tb, CONST_STR_LEN("_plugin_init"));
+		init = (int(WINAPI *)(plugin *))(intptr_t)GetProcAddress(lib, tb->ptr);
+		if (init == NULL) {
+			log_w32_syserror_2(srv->errh, __FILE__, __LINE__,
+			                   "GetProcAddress()", tb->ptr);
+		        FreeLibrary(lib);
+			return -1;
+		}
+	  #else
+	   #if defined(__CYGWIN__)
+		buffer_append_string_len(tb, CONST_STR_LEN(".dll"));
+	   #else
+		buffer_append_string_len(tb, CONST_STR_LEN(".so"));
+	   #endif
+		if (NULL == (lib = dlopen(tb->ptr, RTLD_NOW|RTLD_GLOBAL))) {
 			log_error(srv->errh, __FILE__, __LINE__,
 			  "dlopen() failed for: %s %s", tb->ptr, dlerror());
-
-			plugin_free(p);
-
 			if (srv->srvconf.compat_module_load) {
 				if (buffer_eq_slen(module, CONST_STR_LEN("mod_deflate")))
 					continue;
 			}
 			return -1;
 		}
-
-#endif
 		buffer_clear(tb);
 		buffer_append_str2(tb, BUF_PTR_LEN(module),
                                        CONST_STR_LEN("_plugin_init"));
-
-#ifdef __WIN32
-		init = GetProcAddress(p->lib, tb->ptr);
-
-		if (init == NULL) {
-			LPVOID lpMsgBuf;
-			FormatMessage(
-				FORMAT_MESSAGE_ALLOCATE_BUFFER |
-					FORMAT_MESSAGE_FROM_SYSTEM,
-				NULL,
-				GetLastError(),
-				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-				(LPTSTR) &lpMsgBuf,
-				0, NULL);
-
-			log_error(srv->errh, __FILE__, __LINE__,
-			  "getprocaddress failed: %s %s", tb->ptr, lpMsgBuf);
-
-			plugin_free(p);
-			return -1;
-		}
-
-#else
-#if 1
-		init = (int (*)(plugin *))(intptr_t)dlsym(p->lib, tb->ptr);
-#else
-		*(void **)(&init) = dlsym(p->lib, tb->ptr);
-#endif
+		init = (int (*)(plugin *))(intptr_t)dlsym(lib, tb->ptr);
 		if (NULL == init) {
 			const char *error = dlerror();
 			if (error != NULL) {
@@ -251,26 +230,25 @@ int plugins_load(server *srv) {
 			} else {
 				log_error(srv->errh, __FILE__, __LINE__, "dlsym symbol not found: %s", tb->ptr);
 			}
-
-			plugin_free(p);
+		        dlclose(lib);
 			return -1;
 		}
+	  #endif
+	  }
 
-#endif
+		plugin *p = plugin_init();
+		p->lib = lib;
 		if ((*init)(p)) {
 			log_error(srv->errh, __FILE__, __LINE__, "%s plugin init failed", module->ptr);
-
 			plugin_free(p);
 			return -1;
 		}
-#if 0
-		log_error(srv->errh, __FILE__, __LINE__, "%s plugin loaded", module->ptr);
-#endif
-		plugins_register(srv, p);
+		((plugin **)srv->plugins.ptr)[srv->plugins.used++] = p;
 	}
 
 	return 0;
 }
+
 #endif /* defined(LIGHTTPD_STATIC) */
 
 typedef struct {
@@ -518,8 +496,7 @@ handler_t plugins_call_init(server *srv) {
 	}
 
 	/* allocate and fill slots of two dimensional array */
-	srv->plugin_slots = calloc(nslots, sizeof(plugin_fn_data));
-	force_assert(NULL != srv->plugin_slots);
+	srv->plugin_slots = ck_calloc(nslots, sizeof(plugin_fn_data));
 	memcpy(srv->plugin_slots, offsets, sizeof(offsets));
 
 	/* add handle_uri_raw before handle_uri_clean, but in same slot */
@@ -587,6 +564,5 @@ void plugins_free(server *srv) {
 	free(srv->plugins.ptr);
 	srv->plugins.ptr = NULL;
 	srv->plugins.used = 0;
-	srv->plugins.size = 0;
 	array_free_data(&plugin_stats);
 }
